from typing import TypeAlias

from rainbowadn.core import *
from rainbowadn.flow.bridge import *
from rainbowadn.flow.core import *
from rainbowadn.flow.primitive import *
from rainbowadn.flow.verification.core import *
from rainbowadn.flow.verification.stateverification import *
from rainbowadn.inlining import *
from rainbowadn.nullability import *
from ._flowbank import *
from ._flowcheque import *
from ._flowstandard import *
from ._flowtransaction import *
from ._flowunion import *

__all__ = ('BankFlow',)

Link: TypeAlias = IPair[FlowCheque, FlowBank]


class BankFlow(
    Verification[
        tuple[Nullable[HashPoint[FlowBank]], HashPoint[FlowCheque], HashPoint[FlowBank]]
    ],
):
    def __init__(self, initial: FlowBank):
        assert isinstance(initial, FlowBank)
        self.initial = initial

    @classmethod
    async def _verify_disjoint_union(
            cls,
            per_previous: FlowStandard[FlowCoin],
            per_cheque: FlowStandard[FlowCoin],
            per_bank: FlowStandard[FlowCoin],
    ) -> bool:
        assert isinstance(per_previous, FlowStandard)
        assert isinstance(per_cheque, FlowStandard)
        assert isinstance(per_bank, FlowStandard)

        async def verify_no_intersection():
            assert_true(await per_previous.verify_does_not_intersect(await per_cheque.reducer()))
            return True

        assert_trues(
            await gather(
                per_previous.verify_subset(UnitReducer(per_bank)),
                per_cheque.verify_subset(UnitReducer(per_bank)),
                per_bank.verify_subset(ListBridge([per_previous, per_cheque])),
                verify_no_intersection(),
            )
        )

        return True

    @classmethod
    async def _verify(
            cls,
            previous: FlowBank,
            cheque: FlowCheque,
            bank: FlowBank,
    ) -> bool:
        assert isinstance(previous, FlowBank)
        assert isinstance(cheque, FlowCheque)
        assert isinstance(bank, FlowBank)

        async def verify_unique_minted():
            assert_true(
                await cls._verify_disjoint_union(
                    previous.minted,
                    cheque.minted,
                    bank.minted,
                )
            )
            return True

        async def verify_unique_used():
            assert_true(
                await cls._verify_disjoint_union(
                    previous.used,
                    cheque.used,
                    bank.used,
                )
            )
            return True

        async def verify_used_were_minted():
            assert_true(
                await cheque.used.verify_subset(UnitReducer(bank.minted))
            )
            return True

        assert_trues(
            await gather(
                cheque.verify(),
                verify_unique_minted(),
                verify_unique_used(),
                verify_used_were_minted(),
            )
        )
        return True

    async def verify(
            self,
            element: tuple[Nullable[FlowBank], FlowCheque, FlowBank]
    ) -> bool:
        assert isinstance(element, tuple)
        previous: Nullable[FlowBank]
        cheque: FlowCheque
        bank: FlowBank
        previous, cheque, bank = element
        assert isinstance(previous, Nullable)
        assert isinstance(cheque, FlowCheque)
        assert isinstance(bank, FlowBank)
        previous_bank: FlowBank
        if previous.null():
            previous_bank = self.initial
        else:
            previous_bank = previous.resolve()
        assert isinstance(previous_bank, FlowBank)
        assert_true(await self._verify(previous_bank, cheque, bank))
        return True

    async def add(self, previous: Nullable[FlowBank], cheque: FlowCheque) -> FlowBank:
        assert isinstance(previous, Nullable)
        assert isinstance(cheque, FlowCheque)
        previous_bank: FlowBank
        if previous.null():
            previous_bank = self.initial
        else:
            previous_bank = previous.resolve()
        assert isinstance(previous_bank, FlowBank)
        minted: FlowStandard[FlowCoin]
        used: FlowStandard[FlowCoin]
        minted, used = await gather(
            flow_union(previous_bank.minted, cheque.minted),
            flow_union(previous_bank.used, cheque.used),
        )
        return FlowBank(minted, used)

    def link_verification(self) -> Verification[
        tuple[Nullable[HashPoint[Link]], HashPoint[Link]]
    ]:
        class Decomposition(Mapper[HashPoint[Link], tuple[FlowCheque, FlowBank]]):
            async def map(self, element: HashPoint[Link]) -> tuple[FlowCheque, FlowBank]:
                assert isinstance(element, HashPoint)
                link: Link = await element.resolve()
                assert isinstance(link, IPair)
                cheque: FlowCheque
                bank: FlowBank
                cheque, bank = await gather(
                    link.e0.resolve(),
                    link.e1.resolve(),
                )
                assert isinstance(cheque, FlowCheque)
                assert isinstance(bank, FlowBank)
                return cheque, bank

        return StateVerification(
            Decomposition(),
            self.loose()
        ).loose()

    def loose(self) -> Verification[
        tuple[Nullable[HashPoint[FlowBank]], HashPoint[FlowCheque], HashPoint[FlowBank]]
    ]:
        return self
